#/*
 *
 * Copyright (C) 2011 Christian Brandt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef _SUBGRAPH_ROOT_HPP_DB10380EE37788
#define _SUBGRAPH_ROOT_HPP_DB10380EE37788

#include "graph.hpp"
#include "node.hpp"
#include "edge.hpp"

namespace Gamera { namespace GraphApi {



class SubgraphRoots {
protected:
   struct SubgraphNode {
      Node *n;
      bool root, visited;
      SubgraphNode(Node* n) {
         this->n = n;
         root = visited = false;
      }
   };

   std::stack<SubgraphNode*> nodeStack;
   std::map<Node*, SubgraphNode*> nodeMap;
   std::set<SubgraphNode*> nodeSet;
   size_t i;
   Graph* g;

   void process(SubgraphNode* t);
   
public:
   NodeVector* subgraph_roots(Graph* g);
};



}} // end Gamera::GraphApi
#endif /* _SUBGRAPH_ROOT_HPP_DB10380EE37788 */

