@import url(html4css1.css);
@import url(pygments.css);

body {
  margin: 2em 2em 2em 2em;
  background-color: #effffd;
}

a.toc-backref {
  text-decoration: none ;
  color: black }

h1 {
  background-color: #e1f0ee;
  color: #29493c;
  border-top-color: #72ada8;
  border-top-style: solid;
  border-top-width: 4px
}

h2 {
  background-color: #e1f0ee;
  color: #29493c;
  border-top-color: #72ada8;
  border-top-style: solid;
  border-top-width: 2px
}

h3 {
  background-color: #e1f0ee;
  color: #29493c;
  border-top-color: #72ada8;
  border-top-style: solid;
  border-top-width: 1px
}

h4 {
  background-color: #e1f0ee;
  color: #29493c;
  border-top-color: #72ada8;
  border-top-style: solid;
  border-top-width: 1px
}

h5 {
  background-color: #e1f0ee;
  color: #29493c;
  border-top-color: #72ada8;
  border-top-style: solid;
  border-top-width: 0.5px
}

div.code-block, div.highlight {
margin-left: 2em;
margin-right: 2em;
background-color: #f0f0e0;
font-family: "Andale Mono", "Bitstream Vera Sans Mono", monospace;
border-color: #e0e0d0;
border-style: solid;
border-width: 1px;
font-size: 10pt;
padding: 1em;
}

/* The following is for SilverCity syntax highlighting */


.code_default
{
    FONT-FAMILY: "Andale Mono", "Bitstream Vera Sans Mono", monospace;
    FONT-SIZE: 10pt;
}

.c_character
{
	color: olive;
}

.c_comment
{
	color: green;
	font-style: italic;
}

.c_commentdoc
{
	color: green;
	font-style: italic;
}

.c_commentdockeyword
{
	color: navy;
	font-weight: bold;
}

.c_commentdockeyworderror
{
	color: red;
	font-weight: bold;
}

.c_commentline
{
	color: green;
	font-style: italic;
}

.c_commentlinedoc
{
	color: green;
	font-style: italic;
}

.c_default
{
}

.c_identifier
{
	color: black;
}

.c_number
{
    color: #009999;
}

.c_operator
{
	color: black;
}

.c_preprocessor
{
	color: navy;
	font-weight: bold;
}

.c_regex
{
	color: olive;
}

.c_string
{
	color: olive;
}

.c_stringeol
{
	color: olive;
}

.c_uuid
{
	color: olive;
}

.c_verbatim
{
	color: olive;
}

.c_word
{
	color: navy;
	font-weight: bold;
}

.c_word2
{
	color: navy;
	font-weight: bold;
}

.h_asp
{
    color: #ffff00;
}

.h_aspat
{
    color: #ffdf00;
}

.h_attribute
{
    color: #008080;
}

.h_attributeunknown
{
    color: #ff0000;
}

.h_cdata
{
    color: #ffdf00;
}

.h_comment
{
    color: #808000;
}

.h_default
{
}

.h_doublestring
{
	color: olive;
}

.h_entity
{
    color: #800080;
}

.h_number
{
    color: #009999;
}

.h_other
{
    color: #800080;
}

.h_script
{
    color: #000080;
}

.h_singlestring
{
	color: olive;
}

.h_tag
{
    color: #000080;
}

.h_tagend
{
    color: #000080;
}

.h_tagunknown
{
    color: #ff0000;
}

.h_xmlend
{
    color: #0000ff;
}

.h_xmlstart
{
    color: #0000ff;
}

.pl_array
{
	color: black;
}

.pl_backticks
{
	color: olive;
}

.pl_character
{
	color: olive;
}

.pl_commentline
{
	color: green;
	font-style: italic;
}

.pl_datasection
{
	color: olive;
}

.pl_default
{
}

.pl_error
{
	color: red;
	font-weight: bold;
}

.pl_hash
{
	color: black;
}

.pl_here_delim
{
	color: olive;
}

.pl_here_q
{
	color: olive;
}

.pl_here_qq
{
	color: olive;
}

.pl_here_qx
{
	color: olive;
}

.pl_identifier
{
	color: black;
}

.pl_longquote
{
	color: olive;
}

.pl_number
{
    color: #009999;
}

.pl_operator
{
	color: black;
}

.pl_pod
{
	color: black;
	font-style: italic;
}

.pl_preprocessor
{
	color: navy;
	font-weight: bold;
}

.pl_punctuation
{
	color: black;
}

.pl_regex
{
	color: olive;
}

.pl_regsubst
{
	color: olive;
}

.pl_scalar
{
	color: black;
}

.pl_string
{
	color: olive;
}

.pl_string_q
{
	color: olive;
}

.pl_string_qq
{
	color: olive;
}

.pl_string_qr
{
	color: olive;
}

.pl_string_qw
{
	color: olive;
}

.pl_string_qx
{
	color: olive;
}

.pl_symboltable
{
	color: black;
}

.pl_word
{
	color: navy;
	font-weight: bold;
}

.p_character
{
	color: olive;
}

.p_classname
{
	color: blue;
	font-weight: bold;
}

.p_commentblock
{
	color: gray;
	font-style: italic;
}

.p_commentline
{
	color: green;
	font-style: italic;
}

.p_default
{
}

.p_defname
{
    color: #009999;
	font-weight: bold;
}

.p_identifier
{
	color: black;
}

.p_number
{
    color: #009999;
}

.p_operator
{
	color: black;
}

.p_string
{
	color: olive;
}

.p_stringeol
{
	color: olive;
}

.p_triple
{
	color: olive;
}

.p_tripledouble
{
	color: olive;
}

.p_word
{
	color: navy;
	font-weight: bold;
}

.yaml_comment
{
    color: #008800;
	font-style: italic;
}

.yaml_default
{
}

.yaml_document
{
    color: #808080;
	font-style: italic;
}

.yaml_identifier
{
	color: navy;
	font-weight: bold;
}

.yaml_keyword
{
    color: #880088;
}

.yaml_number
{
    color: #880000;
}

.yaml_reference
{
    color: #008888;
}

