// -*- mode: c++; c-indent-level: 4; c++-member-init-indent: 8; comment-column: 35; -*-

/**
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Contact: http://eodev.sourceforge.net

   Autors: todos@geneura.ugr.es, http://geneura.ugr.es
	   Marc.Schoenauer@polytechnique.fr
	   mak@dhi.dk
	   Caner.Candan@univ-angers.fr
*/

#include <utils/eoSignal.h>

/**
 * @addtogroup Continuators
 * @{
 */

// --- Global variables - but don't know what else to do - MS ---
std::map< int, bool > signals_called;

/** @} */
